import multiprocessing
import os
import shutil
import subprocess
from abc import ABC, abstractmethod
from typing import List, Union

import numpy as np
import scanpy as sc
from gtars.tokenizers import MetaTokenizer as GMetaTokenizer
from gtars.tokenizers import Region as GRegion
from gtars.tokenizers import TreeTokenizer as GTreeTokenizer
from huggingface_hub import hf_hub_download
from rich.progress import track

from geniml.io import Region, RegionSet
from geniml.tokenization.split_file import split_file

from .hard_tokenization_batch import main as hard_tokenization
from .utils import Timer, time_str


class Tokenizer(ABC):
    @abstractmethod
    def tokenize(self, *args, **kwargs):
        raise NotImplementedError


class Namespace:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TreeTokenizer(Tokenizer):
    """
    A fast, in memory, tokenizer that uses `gtokenizers` - a rust based tokenizer.

    This should be used to tokenize bulk data, like BED files. If you need to tokenize single cell data, use
    the `AnnDataTokenizer` instead.
    """

    @classmethod
    def from_pretrained(cls, model_path: str, **kwargs):
        """
        Create a new tokenizer from a pretrained model's vocabulary.

        Usage:
        ```
        tokenizer = TreeTokenizer.from_pretrained("path/to/universe.bed")
        ```

        :param str model_path: The path to the pretrained model on huggingface.
        """
        universe_file_path = hf_hub_download(model_path, "universe.bed")
        return cls(universe_file_path, **kwargs)

    def __init__(self, universe: str) -> GTreeTokenizer:
        """
        Create a new tokenizer.

        This tokenizer only accepts a path to a BED file containing regions.

        :param str universe: The universe to use for tokenization.
        """
        self._tokenizer = GTreeTokenizer(universe)

    @property
    def universe(self):
        return self._tokenizer.universe

    def tokenize(self, query: Union[str, RegionSet]) -> List[List[Region]]:
        """
        Tokenize a Region or RegionSet into the universe

        :param Union[Region, RegionSet] query: The query to tokenize.
        """
        if isinstance(query, sc.AnnData) or isinstance(query, RegionSet):
            result = self._tokenizer(query)
            return result.to_regions()
        else:
            raise ValueError(
                f"Please pass a RegionSet object or a path to a BED file. You passed: {type(query)}"
            )

    def encode(self, query: sc.AnnData) -> List[List[int]]:
        """
        Tokenize an AnnData object to IDs.

        :param sc.AnnData query: The query to tokenize.
        """
        if isinstance(query, sc.AnnData) or isinstance(query, RegionSet):
            result = self._tokenizer(query)
            return result.to_ids()
        else:
            raise ValueError(
                f"Please pass a RegionSet object or a path to a BED file. You passed: {type(query)}"
            )

    def decode(self, query: List[List[int]]) -> List[List[Region]]:
        """
        Decode a list of IDs back to regions.

        :param List[List[int]] query: The query to decode.
        """
        return [self._tokenizer.decode(ids) for ids in query]

    def padding_token(self) -> GRegion:
        return self._tokenizer.padding_token

    def padding_token_id(self) -> int:
        return self._tokenizer.padding_token_id

    def unknown_token(self) -> GRegion:
        return self._tokenizer.unknown_token

    def unknown_token_id(self) -> int:
        return self._tokenizer.unknown_token_id

    def mask_token(self) -> GRegion:
        return self._tokenizer.mask_token

    def mask_token_id(self) -> int:
        return self._tokenizer.mask_token_id

    def cls_token(self) -> GRegion:
        return self._tokenizer.cls_token

    def cls_token_id(self) -> int:
        return self._tokenizer.cls_token_id

    def bos_token(self) -> GRegion:
        return self._tokenizer.bos_token

    def bos_token_id(self) -> int:
        return self._tokenizer.bos_token_id

    def eos_token(self) -> GRegion:
        return self._tokenizer.eos_token

    def eos_token_id(self) -> int:
        return self._tokenizer.eos_token_id

    def sep_token(self) -> GRegion:
        return self._tokenizer.sep_token

    def sep_token_id(self) -> int:
        return self._tokenizer.sep_token_id

    def export(self, path: str):
        return self._tokenizer.export(path)

    def __len__(self):
        return len(self._tokenizer)

    def __call__(self, query: Union[str, RegionSet]) -> List[List[Region]]:
        if isinstance(query, str) or isinstance(query, RegionSet) or isinstance(query, list):
            result = self._tokenizer(query)
            return result
        else:
            raise NotImplementedError(
                "Only RegionSet, region, or str is supported for this tokenizer"
            )

    def __repr__(self):
        return "TreeTokenizer()"

    def __str__(self):
        return "TreeTokenizer()"


class AnnDataTokenizer(Tokenizer):
    """
    A fast, in memory, tokenizer that uses `gtokenizers` - a rust based tokenizer.

    This is actually a wrapper around the core TreeTokenizer. Ideally, we'd tokenize the
    AnnData natively in Rust, but that's a bit more involved. So, we'll just use the
    Python wrapper for now.
    """

    @classmethod
    def from_pretrained(cls, model_path: str, **kwargs):
        """
        Create a new tokenizer from a pretrained model's vocabulary.

        Usage:
        ```
        tokenizer = AnnDataTokenizer.from_pretrained("path/to/universe.bed")
        ```

        :param str model_path: The path to the pretrained model on huggingface.
        """
        universe_file_path = hf_hub_download(model_path, "universe.bed")
        return cls(universe_file_path, **kwargs)

    def __init__(self, universe: str = None, verbose: bool = False, tokenizer_type: str = "tree"):
        """
        Create a new tokenizer.

        This tokenizer only accepts a path to a BED file containing regions.

        :param str universe: The universe to use for tokenization.
        """
        self.verbose = verbose
        self._tokenizer: Union[GTreeTokenizer, GMetaTokenizer]
        if universe is not None:
            if tokenizer_type == "meta":
                self._tokenizer = GMetaTokenizer(universe)
            else:
                self._tokenizer = GTreeTokenizer(universe)
        else:
            self._tokenizer = None

    def _tokenize_anndata(self, adata: sc.AnnData) -> List[List[Region]]:
        """
        Tokenize an AnnData object. This is more involved, so it gets its own function.

        :param sc.AnnData query: The query to tokenize.
        """
        # extract regions from AnnData
        # its weird because of how numpy handle Intervals, the parent class of Region,
        # see here:
        # https://stackoverflow.com/a/43722306/13175187
        adata_features = [
            Region(chr, int(start), int(end))
            for chr, start, end in track(
                zip(adata.var["chr"], adata.var["start"], adata.var["end"]),
                total=adata.var.shape[0],
                description="Extracting regions from AnnData",
                disable=not self.verbose,
            )
        ]
        features = np.ndarray(len(adata_features), dtype=object)
        for i, region in enumerate(adata_features):
            features[i] = region
        del adata_features

        # tokenize
        tokenized = []
        for row in track(
            range(adata.shape[0]),
            total=adata.shape[0],
            description="Tokenizing",
            disable=not self.verbose,
        ):
            _, non_zeros = adata.X[row].nonzero()
            regions = features[non_zeros]
            tokenized.append(self._tokenizer(regions))

        return tokenized

    @property
    def universe(self):
        return self._tokenizer.universe

    def tokenize(self, query: sc.AnnData) -> List[List[Region]]:
        """
        Tokenize a Region or RegionSet into the universe

        :param Union[Region, RegionSet, sc.AnnData] query: The query to tokenize.
        :param bool ids_only: Whether to return only the IDs or the full TokenizedRegionSet
        :param bool as_strings: Whether to return the IDs as strings or ints
        """
        if isinstance(query, sc.AnnData):
            adata = query
        elif isinstance(query, str):
            adata = sc.read_h5ad(query)
        else:
            raise NotImplementedError("Only AnnData is supported right now.")

        adata_features = [
            Region(chr, int(start), int(end))
            for chr, start, end in track(
                zip(adata.var["chr"], adata.var["start"], adata.var["end"]),
                total=adata.var.shape[0],
                description="Extracting regions from AnnData",
                disable=not self.verbose,
            )
        ]
        features = np.ndarray(len(adata_features), dtype=object)
        for i, region in enumerate(adata_features):
            features[i] = region
        del adata_features

        # tokenize
        tokenized = []
        for row in track(
            range(adata.shape[0]),
            total=adata.shape[0],
            description="Tokenizing",
            disable=not self.verbose,
        ):
            _, non_zeros = adata.X[row].nonzero()
            regions = features[non_zeros]
            tokenized.append(self._tokenizer.tokenize(regions))

        return tokenized

    def encode(self, query: Union[sc.AnnData, str]) -> List[List[int]]:
        """
        Tokenize an AnnData object to IDs.

        :param sc.AnnData query: The query to tokenize.
        """
        if isinstance(query, sc.AnnData):
            adata = query
        elif isinstance(query, str):
            adata = sc.read_h5ad(query)
        else:
            raise NotImplementedError("Only AnnData is supported right now.")

        adata_features = [
            Region(chr, int(start), int(end))
            for chr, start, end in track(
                zip(adata.var["chr"], adata.var["start"], adata.var["end"]),
                total=adata.var.shape[0],
                description="Extracting regions from AnnData",
                disable=not self.verbose,
            )
        ]
        features = np.ndarray(len(adata_features), dtype=object)
        for i, region in enumerate(adata_features):
            features[i] = region
        del adata_features

        # tokenize
        tokenized = []
        for row in track(
            range(adata.shape[0]),
            total=adata.shape[0],
            description="Tokenizing",
            disable=not self.verbose,
        ):
            _, non_zeros = adata.X[row].nonzero()
            regions = features[non_zeros]
            tokenized.append(self._tokenizer.encode(regions))

        return tokenized

    def decode(self, query: List[List[int]]) -> List[List[Region]]:
        """
        Decode a list of IDs back to regions.

        :param List[List[int]] query: The query to decode.
        """
        return [self._tokenizer.decode(ids) for ids in query]

    def padding_token(self) -> GRegion:
        return self._tokenizer.padding_token

    def padding_token_id(self) -> int:
        return self._tokenizer.padding_token_id

    def unknown_token(self) -> GRegion:
        return self._tokenizer.unknown_token

    def unknown_token_id(self) -> int:
        return self._tokenizer.unknown_token_id

    def mask_token(self) -> GRegion:
        return self._tokenizer.mask_token

    def mask_token_id(self) -> int:
        return self._tokenizer.mask_token_id

    def cls_token(self) -> GRegion:
        return self._tokenizer.cls_token

    def cls_token_id(self) -> int:
        return self._tokenizer.cls_token_id

    def bos_token(self) -> GRegion:
        return self._tokenizer.bos_token

    def bos_token_id(self) -> int:
        return self._tokenizer.bos_token_id

    def eos_token(self) -> GRegion:
        return self._tokenizer.eos_token

    def eos_token_id(self) -> int:
        return self._tokenizer.eos_token_id

    def sep_token(self) -> GRegion:
        return self._tokenizer.sep_token

    def sep_token_id(self) -> int:
        return self._tokenizer.sep_token_id

    def export(self, path: str):
        return self._tokenizer.export(path)

    def __len__(self):
        return len(self.universe.regions)

    def __call__(self, query: Union[sc.AnnData, str]) -> List[List[Region]]:
        if isinstance(query, sc.AnnData):
            result = self._tokenize_anndata(query)
            return result
        elif isinstance(query, str):
            query = sc.read_h5ad(query)
            result = self._tokenize_anndata(query)
            return result
        else:
            raise NotImplementedError("Only AnnData is supported for this tokenizer.")

    def __repr__(self):
        return "AnnDataTokenizer()"

    def __str__(self):
        return "AnnDataTokenizer()"


def hard_tokenization_main(
    src_folder: str,
    dst_folder: str,
    universe_file: str,
    fraction: float = 1e-9,
    file_list: List[str] = None,
    num_workers: int = 10,
    bedtools_path: str = "bedtools",
) -> int:
    """Tokenizes raw BED files in parallel.

    This is the main function for hard tokenization. It uses multiple processes to
    speed up the tokenization process.

    Args:
        src_folder (str): The folder where raw BED files reside.
        dst_folder (str): The folder to store tokenized BED files.
        universe_file (str): The path to a universe file.
        fraction (float, optional): A parameter for bedtools.intersect.
            Defaults to 1e-9.
        file_list (list[str], optional): A list of files (just names not full
            paths) that need to be tokenized. Defaults to None and uses all BED
            files in src_folder.
        num_workers (int, optional): Number of processes used. Defaults to 10.
        bedtools_path (str, optional): The path to a bedtools binary. Defaults
            to "bedtools".

    Raises:
        Exception: No bedtools executable found

    Returns:
        int: 0 when the dst_folder folder has incomplete list of tokenized BED
            files which should be removed first; 1 when the dst_folder folder
            has the complete tokenized BED files or the tokenization process
            succeeds.
    """
    timer = Timer()
    start_time = timer.t()

    file_list_path = os.path.join(dst_folder, "file_list.txt")
    files = os.listdir(src_folder)
    file_count = len(files)
    if file_count == 0:
        print(f"No files in {src_folder}")
        return 0

    os.makedirs(dst_folder, exist_ok=True)
    if file_list is None:  # use all bed files in data_folder
        # generate a file list
        file_list = files
        print(f"Use all ({file_count}) bed files in {src_folder}")
    else:
        file_number = len(file_list)
        print(f"{file_count} bed files in total, use {file_number} of them")

    # check whether all files in file_list are tokenized
    number = -1
    if os.path.exists(dst_folder):
        all_set = set([f.strip() for f in file_list])
        existing_set = set(os.listdir(dst_folder))
        not_covered = all_set - existing_set
        number = len(not_covered)
    if number == 0 and len(existing_set) == len(all_set):
        print("Skip tokenization. Using the existing tokenization files")
        return 1
    elif len(existing_set) > 0:
        print(
            f"Folder {dst_folder} exists with incomplete tokenized files. Please empty/delete the folder first"
        )
        return 0

    with open(file_list_path, "w") as f:
        for file in file_list:
            f.write(file)
            f.write("\n")

    if bedtools_path == "bedtools":
        try:
            rval = subprocess.call([bedtools_path, "--version"])
        except Exception:
            raise Exception("No bedtools executable found")
        if rval != 0:
            raise Exception("No bedtools executable found")

    print(f"Tokenizing {len(file_list)} bed files ...")

    file_count = len(file_list)
    # split the file_list into several subsets for each worker to process in parallel
    nworkers = min(int(np.ceil(file_count / 20)), num_workers)
    if nworkers <= 1:
        tokenization_args = Namespace(
            data_folder=src_folder,
            file_list=file_list_path,
            token_folder=dst_folder,
            universe=universe_file,
            bedtools_path=bedtools_path,
            fraction=fraction,
        )
        hard_tokenization(tokenization_args)

    else:  # multiprocessing
        dest_folder = os.path.join(dst_folder, "splits")
        split_file(file_list_path, dest_folder, nworkers)
        args_arr = []
        for n in range(nworkers):
            temp_token_folder = os.path.join(dst_folder, f"batch_{n}")
            tokenization_args = Namespace(
                data_folder=src_folder,
                file_list=os.path.join(dest_folder, f"split_{n}.txt"),
                token_folder=temp_token_folder,
                universe=universe_file,
                bedtools_path=bedtools_path,
                fraction=fraction,
            )
            args_arr.append(tokenization_args)
        with multiprocessing.Pool(nworkers) as pool:
            processes = [pool.apply_async(hard_tokenization, args=(param,)) for param in args_arr]
            _ = [r.get() for r in processes]
        # move tokenized files in different folders to expr_tokens
        shutil.rmtree(dest_folder)
        for param in args_arr:
            allfiles = os.listdir(param.token_folder)
            for f in allfiles:
                shutil.move(
                    os.path.join(param.token_folder, f),
                    os.path.join(dst_folder, f),
                )
            shutil.rmtree(param.token_folder)
    os.remove(file_list_path)
    print(f"Tokenization complete {len(os.listdir(dst_folder))}/{file_count} bed files")
    elapsed_time = timer.t() - start_time
    print(f"[Tokenization] {time_str(elapsed_time)}/{time_str(timer.t())}")
    return 1



""" Perturb regions in bedfiles """

import logging
import math
import os
import random
import sys

import logmuse
import numpy as np
import pandas as pd

from geniml.bedshift import BedshiftYAMLHandler, arguments
from geniml.bedshift._version import __version__

_LOGGER = logging.getLogger(__name__)

__all__ = ["Bedshift"]


class Bedshift(object):
    """
    The bedshift object with methods to perturb regions
    """

    def __init__(self, bedfile_path, chrom_sizes=None, delimiter="\t"):
        """
        Read in a .bed file to pandas DataFrame format

        :param str bedfile_path: the path to the BED file
        :param str chrom_sizes: the path to the chrom.sizes file
        :param str delimiter: the delimiter used in the BED file
        """
        self.bedfile_path = bedfile_path
        self.chrom_lens = {}
        if chrom_sizes:
            self._read_chromsizes(chrom_sizes)
        df = self.read_bed(bedfile_path, delimiter=delimiter)
        self.original_num_regions = df.shape[0]
        self.bed = (
            df.astype({0: "object", 1: "int64", 2: "int64", 3: "object"})
            .sort_values([0, 1, 2])
            .reset_index(drop=True)
        )
        self.original_bed = self.bed.copy()

    def _read_chromsizes(self, fp):
        """
        Read chromosome sizes file

        :param str fp: path to the chrom sizes file
        """
        try:
            with open(fp) as f:
                for line in f:
                    line = line.strip().split("\t")
                    chrom = str(line[0])
                    size = int(line[1])
                    self.chrom_lens[chrom] = size
        except FileNotFoundError:
            msg = "Fasta file path {} invalid".format(fp)
            _LOGGER.error(msg)
            raise FileNotFoundError(msg)

        total_len = sum(self.chrom_lens.values())
        self.chrom_weights = [chrom_len / total_len for chrom_len in self.chrom_lens.values()]

    def reset_bed(self):
        """
        Reset the stored bedfile to the state before perturbations
        """
        self.bed = self.original_bed.copy()

    def _precheck(self, rate, requiresChromLens=False, isAdd=False):
        """
        Check if the rate of perturbation is too high or low

        :param float rate: the rate of perturbation
        :param bool requiresChromLens: check if the perturbation requires a chromosome lengths file
        :param bool isAdd: if True, do a special check for the add rate
        """
        if isAdd:
            if rate < 0:
                msg = "Rate must be greater than 0"
                _LOGGER.error(msg)
                raise ValueError(msg)
        else:
            if rate < 0 or rate > 1:
                msg = "Rate must be between 0 and 1"
                _LOGGER.error(msg)
                raise ValueError(msg)
        if requiresChromLens:
            if len(self.chrom_lens) == 0:
                msg = "chrom.sizes file must be specified"
                _LOGGER.error(msg)
                raise FileNotFoundError(msg)

    def pick_random_chroms(self, n):
        """
        Utility function to pick a random chromosome

        :param str n: the number of random chromosomes to pick
        :return str, float chrom_str, chrom_len: chromosome number and length
        """
        chrom_strs = random.choices(list(self.chrom_lens.keys()), weights=self.chrom_weights, k=n)
        chrom_lens = [self.chrom_lens[chrom_str] for chrom_str in chrom_strs]
        return zip(chrom_strs, chrom_lens)

    def add(self, addrate, addmean, addstdev, valid_bed=None, delimiter="\t"):
        """
        Add regions

        :param float addrate: the rate to add regions
        :param float addmean: the mean length of added regions
        :param float addstdev: the standard deviation of the length of added regions
        :param str valid_bed: the file with valid regions where new regions can be added
        :param str delimiter: the delimiter used in valid_bed
        :return int: the number of regions added
        """
        if valid_bed:
            self._precheck(addrate, requiresChromLens=False, isAdd=True)
        else:
            self._precheck(addrate, requiresChromLens=True, isAdd=True)

        rows = self.bed.shape[0]
        num_add = int(rows * addrate)
        new_regions = {0: [], 1: [], 2: [], 3: []}
        if valid_bed:
            valid_regions = self.read_bed(valid_bed, delimiter)
            valid_regions[3] = valid_regions[2] - valid_regions[1]
            total_bp = valid_regions[3].sum()
            valid_regions[4] = valid_regions[3].apply(lambda x: x / total_bp)
            add_rows = random.choices(
                list(range(len(valid_regions))),
                weights=list(valid_regions[4]),
                k=num_add,
            )
            for row in add_rows:
                data = valid_regions.loc[row]
                chrom = data[0]
                start = random.randint(data[1], data[2])
                end = start + int(np.random.normal(addmean, addstdev))
                new_regions[0].append(chrom)
                new_regions[1].append(start)
                new_regions[2].append(end)
                new_regions[3].append("A")
        else:
            random_chroms = self.pick_random_chroms(num_add)
            for chrom_str, chrom_len in random_chroms:
                start = random.randint(1, chrom_len)
                # ensure chromosome length is not exceeded
                end = min(start + int(np.random.normal(addmean, addstdev)), chrom_len)
                new_regions[0].append(chrom_str)
                new_regions[1].append(start)
                new_regions[2].append(end)
                new_regions[3].append("A")
        self.bed = pd.concat([self.bed, pd.DataFrame(new_regions)], ignore_index=True)
        return num_add

    def add_from_file(self, fp, addrate, delimiter="\t"):
        """
        Add regions from another bedfile to this perturbed bedfile

        :param float addrate: the rate to add regions
        :param str fp: the filepath to the other bedfile
        :return int: the number of regions added
        """
        self._precheck(addrate, requiresChromLens=False, isAdd=True)

        rows = self.bed.shape[0]
        num_add = int(rows * addrate)
        df = self.read_bed(fp, delimiter=delimiter)
        dflen = len(df)
        if num_add > dflen:
            _LOGGER.warning(
                "Number of regions to be added ({}) is larger than the provided bedfile size ({}). Adding {} regions.".format(
                    num_add, dflen, dflen
                )
            )
            num_add = dflen
        add_rows = random.sample(list(range(dflen)), num_add)
        add_df = df.loc[add_rows].reset_index(drop=True)
        add_df[3] = pd.Series(["A"] * add_df.shape[0])
        self.bed = pd.concat([self.bed, add_df], ignore_index=True)
        return num_add

    def shift(self, shiftrate, shiftmean, shiftstdev, shift_rows=[]):
        """
        Shift regions

        :param float shiftrate: the rate to shift regions (both the start and end are shifted by the same amount)
        :param float shiftmean: the mean shift distance
        :param float shiftstdev: the standard deviation of the shift distance
        :return int: the number of regions shifted
        """
        self._precheck(shiftrate, requiresChromLens=True)

        rows = self.bed.shape[0]
        if len(shift_rows) == 0:
            shift_rows = random.sample(list(range(rows)), int(rows * shiftrate))
        new_row_list = []
        to_drop = []
        num_shifted = 0
        invalid_shifted = 0
        for row in shift_rows:
            drop_row, new_region = self._shift(
                row, shiftmean, shiftstdev
            )  # shifted rows display a 1
            if drop_row is not None and new_region:
                num_shifted += 1
                new_row_list.append(new_region)
                to_drop.append(drop_row)
            else:
                invalid_shifted += 1
        self.bed = self.bed.drop(to_drop)
        self.bed = pd.concat([self.bed, pd.DataFrame(new_row_list)], ignore_index=True)
        self.bed = self.bed.reset_index(drop=True)
        if invalid_shifted > 0:
            _LOGGER.warning(
                f"{invalid_shifted} regions were prevented from being shifted outside of chromosome boundaries. Reported regions shifted will be less than expected."
            )
        return num_shifted

    def _shift(self, row, mean, stdev):
        theshift = int(np.random.normal(mean, stdev))

        chrom = self.bed.loc[row][0]
        start = self.bed.loc[row][1]
        end = self.bed.loc[row][2]
        if start + theshift < 0 or end + theshift > self.chrom_lens[str(chrom)]:
            # check if the region is shifted out of chromosome length bounds
            return None, None

        return row, {0: chrom, 1: start + theshift, 2: end + theshift, 3: "S"}

    def shift_from_file(self, fp, shiftrate, shiftmean, shiftstdev, delimiter="\t"):
        """
        Shift regions that overlap the specified file's regions

        :param str fp: the file on which to find overlaps
        :param float shiftrate: the rate to shift regions (both the start and end are shifted by the same amount)
        :param float shiftmean: the mean shift distance
        :param float shiftstdev: the standard deviation of the shift distance
        :param str delimiter: the delimiter used in fp
        :return int: the number of regions shifted
        """
        self._precheck(shiftrate, requiresChromLens=True)

        rows = self.bed.shape[0]
        num_shift = int(rows * shiftrate)

        intersect_regions = self._find_overlap(fp)
        original_colnames = self.bed.columns
        intersect_regions.columns = [str(col) for col in intersect_regions.columns]
        self.bed.columns = [str(col) for col in self.bed.columns]
        indices_of_overlap_regions = self.bed.reset_index().merge(intersect_regions)["index"]
        self.bed.columns = [int(col) for col in self.bed.columns]

        interlen = len(indices_of_overlap_regions)
        if num_shift > interlen:
            _LOGGER.warning(
                "Desired regions shifted ({}) is greater than the number of overlaps found ({}). Shifting {} regions.".format(
                    num_shift, interlen, interlen
                )
            )
            num_shift = len(indices_of_overlap_regions)

        elif interlen > num_shift:
            indices_of_overlap_regions = indices_of_overlap_regions.sample(n=num_shift)

        indices_of_overlap_regions = indices_of_overlap_regions.to_list()

        return self.shift(shiftrate, shiftmean, shiftstdev, indices_of_overlap_regions)

    def cut(self, cutrate):
        """
        Cut regions to create two new regions

        :param float cutrate: the rate to cut regions into two separate regions
        :return int: the number of regions cut
        """
        self._precheck(cutrate)

        rows = self.bed.shape[0]
        cut_rows = random.sample(list(range(rows)), int(rows * cutrate))
        new_row_list = []
        to_drop = []
        for row in cut_rows:
            drop_row, new_regions = self._cut(row)  # cut rows display a 2
            new_row_list.extend(new_regions)
            to_drop.append(drop_row)
        self.bed = self.bed.drop(to_drop)
        self.bed = pd.concat([self.bed, pd.DataFrame(new_row_list)], ignore_index=True)
        self.bed = self.bed.reset_index(drop=True)
        return len(cut_rows)

    def _cut(self, row):
        chrom = self.bed.loc[row][0]
        start = self.bed.loc[row][1]
        end = self.bed.loc[row][2]

        # choose where to cut the region
        thecut = (start + end) // 2  # int(np.random.normal((start+end)/2, (end - start)/6))
        if thecut <= start:
            thecut = start + 10
        if thecut >= end:
            thecut = end - 10

        """ may add in later, this makes the api confusing!
        # adjust the cut regions using the shift function
        new_segs = self.__shift(new_segs, 0, meanshift, stdevshift)
        new_segs = self.__shift(new_segs, 1, meanshift, stdevshift)
        """

        return (
            row,
            [
                {0: chrom, 1: start, 2: thecut, 3: "C"},
                {0: chrom, 1: thecut, 2: end, 3: "C"},
            ],
        )

    def merge(self, mergerate):
        """
        Merge two regions into one new region

        :param float mergerate: the rate to merge two regions into one
        :return int: number of regions merged
        """
        self._precheck(mergerate)

        rows = self.bed.shape[0]
        merge_rows = random.sample(list(range(rows)), int(rows * mergerate))
        to_add = []
        to_drop = []
        for row in merge_rows:
            drop_rows, add_row = self._merge(row)
            if drop_rows and add_row:
                to_add.append(add_row)
                to_drop.extend(drop_rows)
        self.bed = self.bed.drop(to_drop)
        self.bed = pd.concat([self.bed, pd.DataFrame(to_add)], ignore_index=True)
        self.bed = self.bed.reset_index(drop=True)
        return len(to_drop)

    def _merge(self, row):
        # check if the regions being merged are on the same chromosome
        if row + 1 not in self.bed.index or self.bed.loc[row][0] != self.bed.loc[row + 1][0]:
            return None, None

        chrom = self.bed.loc[row][0]
        start = self.bed.loc[row][1]
        end = self.bed.loc[row + 1][2]
        return [row, row + 1], {0: chrom, 1: start, 2: end, 3: "M"}

    def drop(self, droprate):
        """
        Drop regions

        :param float droprate: the rate to drop/remove regions
        :return int: the number of rows dropped
        """
        self._precheck(droprate)

        rows = self.bed.shape[0]
        drop_rows = random.sample(list(range(rows)), int(rows * droprate))
        self.bed = self.bed.drop(drop_rows)
        self.bed = self.bed.reset_index(drop=True)
        return len(drop_rows)

    def drop_from_file(self, fp, droprate, delimiter="\t"):
        """
        drop regions that overlap between the reference bedfile and the provided bedfile.

        :param float droprate: the rate to drop regions
        :param str fp: the filepath to the other bedfile containing regions to be dropped
        :return int: the number of regions dropped
        """
        self._precheck(droprate)

        rows = self.bed.shape[0]
        num_drop = int(rows * droprate)
        drop_bed = self.read_bed(fp, delimiter=delimiter)

        intersect_regions = self._find_overlap(drop_bed)
        original_colnames = self.bed.columns
        intersect_regions.columns = [str(col) for col in intersect_regions.columns]
        self.bed.columns = [str(col) for col in self.bed.columns]
        indices_of_overlap_regions = self.bed.reset_index().merge(intersect_regions)["index"]
        self.bed.columns = [int(col) for col in self.bed.columns]

        interlen = len(indices_of_overlap_regions)
        if num_drop > interlen:
            _LOGGER.warning(
                "Desired regions dropped ({}) is greater than the number of overlaps found ({}). Dropping {} regions.".format(
                    num_drop, interlen, interlen
                )
            )
            num_drop = len(indices_of_overlap_regions)
        elif interlen > num_drop:
            indices_of_overlap_regions = indices_of_overlap_regions.sample(n=num_drop)
        indices_of_overlap_regions = indices_of_overlap_regions.to_list()

        self.bed = self.bed.drop(indices_of_overlap_regions)
        return num_drop

    def set_seed(self, seednum):
        try:
            seednum = int(seednum)
            random.seed(seednum)
            np.random.seed(seednum)
        except ValueError:
            msg = "Seed should be an integer, not {}.".format(type(seednum))
            _LOGGER.error(msg)
            raise ValueError(msg)

    def _find_overlap(self, fp, reference=None):
        """
        Find intersecting regions between the reference bedfile and the comparison file provided in the yaml config file.

        :param str fp: path to file, or pandas DataFrame, for comparison
        :param str reference: path to file, or pandas DataFrame, for reference. If None, then defaults to the original BED file provided to the Bedshift constructor
        :return pd.DataFrame: a DataFrame of overlapping regions
        """
        if reference is None:
            reference_bed = self.original_bed.copy()
        else:
            if isinstance(reference, pd.DataFrame):
                reference_bed = reference.copy()
            elif isinstance(reference, str):
                reference_bed = self.read_bed(reference)
            else:
                raise Exception("unsupported input type: {}".format(type(reference)))
        if isinstance(fp, pd.DataFrame):
            comparison_bed = fp.copy()
        elif isinstance(fp, str):
            comparison_bed = self.read_bed(fp)
        else:
            raise Exception("unsupported input type: {}".format(type(reference)))
        reference_bed.columns = ["Chromosome", "Start", "End", "modifications"]
        comparison_bed.columns = ["Chromosome", "Start", "End", "modifications"]
        # TODO, switch this overlap calculation to use genomicranges
        raise NotImplementedError(
            "This relies on pyranges, which was removed with the switch to geniml."
        )

        # USE AILIST HERE:
        reference_pr = pr.PyRanges(reference_bed)
        comparison_pr = pr.PyRanges(comparison_bed)
        intersection = reference_pr.overlap(comparison_pr, how="first").as_df()
        if len(intersection) == 0:
            raise Exception(
                "no intersection found between {} and {}".format(reference_bed, comparison_bed)
            )
        intersection = intersection.drop(["modifications"], axis=1)
        intersection.columns = [0, 1, 2]
        return intersection

    def all_perturbations(
        self,
        addrate=0.0,
        addmean=320.0,
        addstdev=30.0,
        addfile=None,
        valid_regions=None,
        shiftrate=0.0,
        shiftmean=0.0,
        shiftstdev=150.0,
        shiftfile=None,
        cutrate=0.0,
        mergerate=0.0,
        droprate=0.0,
        dropfile=None,
        yaml=None,
        seed=None,
    ):
        """
        Perform all five perturbations in the order of shift, add, cut, merge, drop.

        :param float addrate: the rate (as a proportion of the total number of regions) to add regions
        :param float addmean: the mean length of added regions
        :param float addstdev: the standard deviation of the length of added regions
        :param str addfile: the file containing regions to be added
        :param str valid_regions: the file containing regions where new regions can be added
        :param float shiftrate: the rate to shift regions (both the start and end are shifted by the same amount)
        :param float shiftmean: the mean shift distance
        :param float shiftstdev: the standard deviation of the shift distance
        :param str shiftfile: the file containing regions to be shifted
        :param float cutrate: the rate to cut regions into two separate regions
        :param float mergerate: the rate to merge two regions into one
        :param float droprate: the rate to drop/remove regions
        :param str dropfile: the file containing regions to be dropped
        :param str yaml: the yaml_config filepath
        :param bedshift.Bedshift bedshifter: Bedshift instance
        :param int seed: a seed for allowing reproducible perturbations
        :return int: the number of total regions perturbed
        """
        if seed:
            self.set_seed(seed)
        if yaml:
            return BedshiftYAMLHandler.BedshiftYAMLHandler(self, yaml).handle_yaml()
        n = 0
        if shiftrate > 0:
            if shiftfile:
                n += self.shift_from_file(shiftfile, shiftrate, shiftmean, shiftstdev)
            else:
                n += self.shift(shiftrate, shiftmean, shiftstdev)
        if addrate > 0:
            if addfile:
                n += self.add_from_file(addfile, addrate)
            else:
                n += self.add(addrate, addmean, addstdev, valid_regions)
        if cutrate > 0:
            n += self.cut(cutrate)
        if mergerate > 0:
            n += self.merge(mergerate)
        if droprate > 0:
            if dropfile:
                n += self.drop_from_file(dropfile, droprate)
            else:
                n += self.drop(droprate)

        return n

    def to_bed(self, outfile_name):
        """
        Write a pandas dataframe back into BED file format

        :param str outfile_name: The name of the output BED file
        """
        self.bed.sort_values([0, 1, 2], inplace=True)
        self.bed.to_csv(outfile_name, sep="\t", header=False, index=False, float_format="%.0f")

    def read_bed(self, bedfile_path, delimiter="\t"):
        """
        Read a BED file into pandas dataframe

        :param str bedfile_path: The path to the BED file
        """
        try:
            df = pd.read_csv(
                bedfile_path,
                sep=delimiter,
                header=None,
                usecols=[0, 1, 2],
                engine="python",
            )
        except FileNotFoundError:
            msg = "BED file path {} invalid".format(bedfile_path)
            _LOGGER.error(msg)
            raise FileNotFoundError(msg)
        except:
            msg = "File {} could not be read".format(bedfile_path)
            _LOGGER.error(msg)
            raise Exception(msg)

        # if there is a header line in the table, remove it
        if not str(df.iloc[0, 1]).isdigit():
            df = df[1:].reset_index(drop=True)

        df[3] = "-"  # column indicating which modifications were made
        return df


def main():
    """Primary workflow"""

    parser = logmuse.add_logging_options(arguments.build_argparser())
    args, remaining_args = parser.parse_known_args()
    global _LOGGER
    _LOGGER = logmuse.logger_via_cli(args)

    _LOGGER.info("Welcome to bedshift version {}".format(__version__))
    _LOGGER.info("Shifting file: '{}'".format(args.bedfile))

    if not args.bedfile:
        parser.print_help()
        msg = "No BED file given"
        _LOGGER.error(msg)
        raise MissingArgumentError(msg)

    if args.chrom_lengths:
        pass
    elif args.genome:
        try:
            import refgenconf

            rgc = refgenconf.RefGenConf(refgenconf.select_genome_config())
            args.chrom_lengths = rgc.seek(args.genome, "fasta", None, "chrom_sizes")
        except ModuleNotFoundError:
            msg = "You must have package refgenconf installed to use a refgenie genome"
            _LOGGER.error(msg)
            raise ModuleNotFoundError(msg)

    msg = arguments.param_msg

    if args.repeat < 1:
        msg = "Repeats must be greater than 0"
        _LOGGER.error(msg)
        raise ValueError(msg)

    if args.outputfile:
        outfile_base = args.outputfile
    else:
        outfile_base = "bedshifted_{}".format(os.path.basename(args.bedfile))

    _LOGGER.info(
        msg.format(
            bedfile=args.bedfile,
            chromsizes=args.chrom_lengths,
            droprate=args.droprate,
            dropfile=args.dropfile,
            addrate=args.addrate,
            addmean=args.addmean,
            addstdev=args.addstdev,
            addfile=args.addfile,
            valid_regions=args.valid_regions,
            shiftrate=args.shiftrate,
            shiftmean=args.shiftmean,
            shiftstdev=args.shiftstdev,
            shiftfile=args.shiftfile,
            cutrate=args.cutrate,
            mergerate=args.mergerate,
            outputfile=outfile_base,
            repeat=args.repeat,
            yaml_config=args.yaml_config,
            seed=args.seed,
        )
    )

    bedshifter = Bedshift(args.bedfile, args.chrom_lengths)
    _LOGGER.info(f"Generating {args.repeat} repetitions...")

    pct_reports = [int(x * args.repeat / 100) for x in [5, 25, 50, 75, 100]]

    for i in range(args.repeat):
        n = bedshifter.all_perturbations(
            args.addrate,
            args.addmean,
            args.addstdev,
            args.addfile,
            args.valid_regions,
            args.shiftrate,
            args.shiftmean,
            args.shiftstdev,
            args.shiftfile,
            args.cutrate,
            args.mergerate,
            args.droprate,
            args.dropfile,
            args.yaml_config,
            args.seed,
        )
        if args.repeat == 1:
            bedshifter.to_bed(outfile_base)
            _LOGGER.info(
                "REGION COUNT | original: {}\tnew: {}\tchanged: {}\t\noutput file: {}".format(
                    bedshifter.original_num_regions,
                    bedshifter.bed.shape[0],
                    str(n),
                    outfile_base,
                )
            )
        else:
            basename, ext = os.path.splitext(os.path.basename(outfile_base))
            dirname = os.path.dirname(outfile_base)
            digits = int(math.log10(args.repeat)) + 1

            rep = str(i + 1).zfill(digits)
            modified_outfile_path = os.path.join(dirname, f"{basename}_rep{rep}{ext}")
            bedshifter.to_bed(modified_outfile_path)

            pct_finished = int((100 * (i + 1)) / args.repeat)
            if i + 1 in pct_reports:
                _LOGGER.info(
                    f"Rep {i+1}. Finished: {pct_finished}%. Output file: {modified_outfile_path}"
                )

        bedshifter.reset_bed()


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        _LOGGER.error("Program canceled by user!")
        sys.exit(1)



def build_subparser_gdst(parser):
    """
    Builds argument parser to support the gdst command line interface (under eval).
    """

    parser.add_argument(
        "--model-path",
        required=True,
        type=str,
        help="path to a Region2Vec model or a Base model",
    )
    parser.add_argument("--embed-type", required=True, type=str, help="region2vec or base")
    parser.add_argument(
        "--num-samples",
        default=10000,
        type=int,
        help="number of samples used in calculation",
    )
    parser.add_argument(
        "--seed",
        default=42,
        type=int,
        help="random seed",
    )

    return parser


def build_subparser_npt(parser):
    """
    Builds argument parser to support the npt command line interface  (under eval).
    """
    parser.add_argument(
        "--model-path",
        required=True,
        type=str,
        help="path to a region2vec model or a Base model",
    )
    parser.add_argument("--embed-type", required=True, type=str, help="region2vec or base")
    parser.add_argument("--K", required=True, type=int, help="number of nearest regions")
    parser.add_argument(
        "--num-samples",
        default=1000,
        type=int,
        help="number of samples used in calculation",
    )
    parser.add_argument(
        "--seed",
        default=42,
        type=int,
        help="random seed",
    )
    parser.add_argument(
        "--num-workers",
        default=10,
        type=int,
        help="number of parllel processes",
    )
    return parser


def build_subparser_ctt(parser):
    """
    Builds argument parser to support the ctt command line interface  (under eval).
    """
    parser.add_argument(
        "--model-path",
        required=True,
        type=str,
        help="path to a region2vec model or a Base model",
    )
    parser.add_argument("--embed-type", required=True, type=str, help="region2vec or base")

    parser.add_argument(
        "--num-samples",
        default=10000,
        type=int,
        help="number of samples used in calculation",
    )
    parser.add_argument(
        "--seed",
        default=42,
        type=int,
        help="random seed",
    )
    parser.add_argument(
        "--num-workers",
        default=10,
        type=int,
        help="number of parllel processes",
    )

    return parser


def build_subparser_rct(parser):
    """
    Builds argument parser to support the rct command line interface  (under eval).
    """
    parser.add_argument(
        "--model-path",
        required=True,
        type=str,
        help="path to a region2vec model or a Base model",
    )
    parser.add_argument(
        "--bin-path",
        required=True,
        type=str,
        help="path to a set of binary embedding",
    )
    parser.add_argument(
        "--embed-type",
        required=True,
        type=str,
        help="region2vec or base for model-path",
    )

    parser.add_argument(
        "--cv-num",
        default=5,
        type=int,
        help="number of folds in cross-validation",
    )
    parser.add_argument(
        "--seed",
        default=42,
        type=int,
        help="random seed",
    )
    parser.add_argument(
        "--out-dim",
        default=-1,
        type=int,
        help="Used when the binary embeddings are very high-dimensional, i.e., there are many training files. Default -1 represents using all the dimensions",
    )
    parser.add_argument(
        "--num-workers",
        default=10,
        type=int,
        help="number of parllel processes",
    )

    return parser


def build_subparser_bingen(parser):
    """
    Builds argument parser to support the bin-gen command line interface  (under eval).
    """
    parser.add_argument(
        "--universe",
        required=True,
        type=str,
        help="path to a universe file",
    )
    parser.add_argument(
        "--token-folder",
        required=True,
        type=str,
        help="path to a folder storing tokenized files",
    )
    parser.add_argument(
        "--file-name",
        required=True,
        type=str,
        help="name of the generated binary embeddings",
    )

    return parser


def build_subparser(parser):
    """
    Builds argument parser to support the eval command line interface.
    """
    sp = parser.add_subparsers(dest="subcommand")
    msg_by_cmd = {
        "gdst": "Genome distance scaling test",
        "npt": "Neighborhood preserving test",
        "ctt": "Cluster tendency test",
        "rct": "Reconstruction test",
        "bin-gen": "Generate binary embeddings",
    }
    subparsers = {}
    for k, v in msg_by_cmd.items():
        subparsers[k] = sp.add_parser(k, description=v, help=v)
    subparsers["gdst"] = build_subparser_gdst(subparsers["gdst"])
    subparsers["npt"] = build_subparser_npt(subparsers["npt"])
    subparsers["ctt"] = build_subparser_ctt(subparsers["ctt"])
    subparsers["rct"] = build_subparser_rct(subparsers["rct"])
    subparsers["bin-gen"] = build_subparser_bingen(subparsers["bin-gen"])
    return parser



import gzip
import logging
import os
from hashlib import md5
from typing import List, NoReturn, Union

import genomicranges
import numpy as np
import pandas as pd
from iranges import IRanges
from ubiquerg import is_url

from .const import (
    MAF_CENTER_COL_NAME,
    MAF_CHROMOSOME_COL_NAME,
    MAF_END_COL_NAME,
    MAF_ENTREZ_GENE_ID_COL_NAME,
    MAF_FILE_DELIM,
    MAF_HUGO_SYMBOL_COL_NAME,
    MAF_NCBI_BUILD_COL_NAME,
    MAF_START_COL_NAME,
    MAF_STRAND_COL_NAME,
)
from .exceptions import BackedFileNotAvailableError, BEDFileReadError
from .utils import compute_md5sum_bedset, extract_maf_col_positions, is_gzipped, read_bedset_file

_LOGGER = logging.getLogger("bbclient")


class Region:
    def __init__(self, chr: str, start: int, stop: int):
        """
        Instantiate a Region object.

        :param str chr: chromosome
        :param int start: start position
        :param int stop: stop position
        """
        self.chr = chr
        self.start = start
        self.end = stop

    def __repr__(self):
        return f"Region({self.chr}, {self.start}, {self.end})"


class RegionSet:
    def __init__(self, regions: Union[str, List[Region]], backed: bool = False):
        """
        Instantiate a RegionSet object. This can be backed or not backed. It represents a set of genomic regions.

        If you specify `backed` as True, then the bed file will not be loaded into memory. This is useful for large
        bed files. You can still iterate over the regions, but you cannot index into them.

        :param regions: path, or url to bed file or list of Region objects
        :param backed: whether to load the bed file into memory or not [Default: False]
        """
        self._df: Union[pd.DataFrame, None] = None

        if isinstance(regions, str):
            self.backed = backed
            self.regions: List[Region] = []
            self.path = regions
            self.regions = None
            self.is_gzipped = False

            if backed:
                if is_url(regions):
                    raise BackedFileNotAvailableError()
                # Open function depending on file type
                if not is_gzipped(regions):
                    open_func = open
                    mode = "r"
                else:
                    self.is_gzipped = True
                    open_func = gzip.open
                    mode = "rt"

                # https://stackoverflow.com/a/32607817/13175187
                try:
                    with open_func(self.path, mode) as file:
                        self.length = sum(1 for line in file if line.strip())
                except UnicodeDecodeError:
                    self.is_gzipped = True
                    with gzip.open(self.path, "rt") as file:
                        self.length = sum(1 for line in file if line.strip())

            else:
                if is_gzipped(regions):
                    df = self._read_gzipped_file(regions)
                else:
                    df = self._read_file_pd(
                        regions,
                        sep="\t",
                        header=None,
                        engine="pyarrow",
                    )
                self._df = df

                _regions = []
                df.apply(
                    lambda row: _regions.append(Region(row[0], row[1], row[2])),
                    axis=1,
                )

                self.regions = _regions
                self.length = len(self.regions)

        # load from list
        elif isinstance(regions, list) and all([isinstance(region, Region) for region in regions]):
            self.backed = False
            self.path = None
            self.regions = regions
            self.length = len(self.regions)
        else:
            raise ValueError("regions must be a path to a bed file or a list of Region objects")

        self._identifier = None

    def to_pandas(self) -> Union[pd.DataFrame, None]:
        if self._df is None:
            seqnames, starts, ends = zip(
                *[(region.chr, region.start, region.end) for region in self]
            )
            return pd.DataFrame([seqnames, starts, ends])

        return self._df

    def _read_gzipped_file(self, file_path: str) -> pd.DataFrame:
        """
        Read a gzipped file into a pandas dataframe

        :param file_path: path to gzipped file
        :return: pandas dataframe
        """
        return self._read_file_pd(
            file_path,
            sep="\t",
            compression="gzip",
            header=None,
            # engine="pyarrow",
        )

    def _read_file_pd(self, *args, **kwargs) -> pd.DataFrame:
        """
        Read bed file into a pandas DataFrame, and skip header rows if needed

        :return: pandas dataframe
        """
        max_rows = 5
        row_count = 0
        while row_count <= max_rows:
            try:
                df = pd.read_csv(*args, **kwargs, skiprows=row_count)
                if row_count > 0:
                    _LOGGER.info(f"Skipped {row_count} rows while standardization. File: '{args}'")
                df = df.dropna(axis=1)
                for index, row in df.iterrows():
                    if (
                        isinstance(row[0], str)
                        and isinstance(row[1], int)
                        and isinstance(row[2], int)
                    ):
                        return df
                    else:
                        if isinstance(row[1], str):
                            try:
                                _ = int(row[1])
                                df[1] = pd.to_numeric(df[1])
                            except ValueError:
                                row_count += 1
                                break
                        if isinstance(row[2], str):
                            try:
                                _ = int(row[2])
                                df[2] = pd.to_numeric(df[2])
                            except ValueError:
                                row_count += 1
                                break
                        return df
            except (pd.errors.ParserError, pd.errors.EmptyDataError) as _:
                if row_count <= max_rows:
                    row_count += 1
        raise BEDFileReadError("Cannot read bed file.")

    def __len__(self):
        return self.length

    def __getitem__(self, key):
        if self.backed:
            raise NotImplementedError("Backed RegionSets do not currently support indexing.")
        else:
            return self.regions[key]

    def __repr__(self):
        if self.path:
            if self.backed:
                return f"RegionSet({self.path}, backed=True)"
            else:
                return f"RegionSet({self.path})"
        else:
            return f"RegionSet(n={self.length})"

    def __iter__(self):
        if self.backed:
            # Open function depending on file type
            if self.is_gzipped:
                open_func = gzip.open
                mode = "rt"
            else:
                open_func = open
                mode = "r"

            with open_func(self.path, mode) as f:
                skipped_lines = 0
                max_skipped_lines = 5
                for line in f:
                    try:
                        chr, start, stop = line.split("\t")[:3]
                    except ValueError as _:
                        if skipped_lines < max_skipped_lines:
                            skipped_lines += 1
                            continue
                        else:
                            raise BEDFileReadError("Could not read line bed file")
                    if skipped_lines > 0:
                        _LOGGER.info(
                            f"Skipped {skipped_lines} lines while opening file. File: '{self.path}'"
                        )
                    yield Region(chr, int(start), int(stop))
        else:
            for region in self.regions:
                yield region

    @property
    def identifier(self) -> str:
        return self.compute_bed_identifier()

    def to_granges(self) -> genomicranges.GenomicRanges:
        """
        Return GenomicRanges contained in this BED file

        :return: GenomicRanges object
        """

        seqnames, starts, ends = zip(
            *[(region.chr, region.start, region.end) for region in self.regions]
        )
        width_list = []
        for start, end in zip(starts, ends):
            width_list.append(end - start)
        ir = IRanges(start=starts, width=width_list)

        return genomicranges.GenomicRanges(seqnames, ir)

    def compute_bed_identifier(self) -> str:
        """
        Return bed file identifier. If it is not set, compute one

        :return: the identifier of BED file (str)
        """
        if self._identifier is not None:
            return self._identifier
        else:
            if not self.backed:
                # concate column values
                chrs = ",".join([region.chr for region in self.regions])
                starts = ",".join([str(region.start) for region in self.regions])
                ends = ",".join([str(region.end) for region in self.regions])

            else:
                open_func = open if not is_gzipped(self.path) else gzip.open
                mode = "r" if not is_gzipped(self.path) else "rt"
                with open_func(self.path, mode) as f:
                    # concate column values
                    chrs = []
                    starts = []
                    ends = []
                    for row in f:
                        chrs.append(row.split("\t")[0])
                        starts.append(row.split("\t")[1])
                        ends.append(row.split("\t")[2].replace("\n", ""))
                    chrs = ",".join(chrs)
                    starts = ",".join(starts)
                    ends = ",".join(ends)

            # hash column values
            chr_digest = md5(chrs.encode("utf-8")).hexdigest()
            start_digest = md5(starts.encode("utf-8")).hexdigest()
            end_digest = md5(ends.encode("utf-8")).hexdigest()
            # hash column digests
            bed_digest = md5(
                ",".join([chr_digest, start_digest, end_digest]).encode("utf-8")
            ).hexdigest()

            self._identifier = bed_digest

            return self._identifier


class BedSet:
    """
    BedSet object
    """

    def __init__(
        self,
        region_sets: Union[List[RegionSet], List[str], List[List[Region]], None] = None,
        file_path: str = None,
        identifier: str = None,
    ):
        """
        :param region_sets: list of BED file paths, RegionSet, or 2-dimension list of Region [Default: None - empty BedSet]
        :param file_path: path to the .txt file with identifier of all BED files in it
        :param identifier: the identifier of the BED set
        """

        if isinstance(region_sets, list):
            # init with a list of BED files
            if all(isinstance(region_set, RegionSet) for region_set in region_sets):
                self.region_sets = region_sets
            # init with a list of file paths or a 2d list of Region
            else:
                self.region_sets = []
                for r in region_sets:
                    self.region_sets.append(RegionSet(r))

        elif file_path is not None:
            if os.path.isfile(file_path):
                self.region_sets = [RegionSet(r) for r in read_bedset_file(file_path)]
            else:
                raise FileNotFoundError(f"The specified file '{file_path}' does not exist.")
        else:
            # create empty regionSet
            self.region_sets = []

        self._bedset_identifier = identifier

    def __len__(self):
        return len(self.region_sets)

    def __iter__(self):
        for region_set in self.region_sets:
            yield region_set

    def __getitem__(self, indx: int):
        return self.region_sets[indx]

    @property
    def identifier(self) -> str:
        return self._bedset_identifier or self.compute_bedset_identifier()

    def add(self, bedfile: RegionSet) -> NoReturn:
        """
        Add a BED file to the BED set

        !Warning: if new bedfile will be added, bedSet identifier will be changed!

        :param bedfile: RegionSet instance, that should be added to the bedSet
        :return: NoReturn
        """
        self.region_sets.append(bedfile)

        self._bedset_identifier = self.compute_bedset_identifier()

    def to_granges_list(self) -> genomicranges.GenomicRangesList:
        """
        Process a list of BED set identifiers and returns a GenomicRangesList object
        """
        gr_list = []
        for regionset in self.region_sets:
            gr_list.append(regionset.to_granges())

        return genomicranges.GenomicRangesList(ranges=gr_list)

    def compute_bedset_identifier(self) -> str:
        """
        Return the identifier. If it is not set, compute one

        :param bedset: BedSet object
        :return: the identifier of BED set
        """
        if self._bedset_identifier is not None:
            return self._bedset_identifier

        elif self._bedset_identifier is None:
            bedfile_ids = []
            for bedfile in self.region_sets:
                bedfile_ids.append(bedfile.compute_bed_identifier())
            self._bedset_identifier = compute_md5sum_bedset(bedfile_ids)

            return self._bedset_identifier


class SNP:
    """
    Python representation of a SNP
    """

    def __init__(
        self,
        hugo_symbol: str = None,
        entrez_gene_id: str = None,
        center: str = None,
        ncbi_build: str = None,
        chromosome: str = None,
        start_position: int = None,
        end_position: int = None,
        strand: str = None,
    ):
        self.hugo_symbol = hugo_symbol
        self.entrez_gene_id = entrez_gene_id
        self.center = center
        self.ncbi_build = ncbi_build
        self.chromosome = chromosome
        self.start_position = start_position
        self.end_position = end_position
        self.strand = strand

    @property
    def start(self):
        return self.start_position

    @property
    def end(self):
        return self.end_position

    @property
    def chr(self):
        return self.chromosome

    def to_region(self):
        chr = self.chromosome
        start = int(self.start_position)
        end = int(self.end_position)
        # bump end position by 1 if needed
        if start == end:
            end += 1

        return Region(chr, start, end)

    def __len__(self):
        return self.end - self.start

    def __repr__(self):
        return f"SNP({self.chromosome}, {self.start_position}, {self.end_position}, {self.strand})"


class Maf:
    """
    Python representation of a MAF file, only supports some columns for now
    """

    def _extract_value_from_col(self, col_name: str, line: str) -> any:
        """
        Extract a value from a column in a line of a MAF file.

        :param col_name: name of column
        :param line: line from MAF file
        :return: value of column
        """
        return line[self.col_positions[col_name]] if self.col_positions[col_name] else None

    def __init__(
        self,
        maf_file: str,
        backed: bool = False,
        bump_end_position: bool = False,
        chr_rep_as_int: bool = False,
    ):
        """
        :param maf_file: path to maf file
        :param backed: whether to load the maf file into memory or not
        :param bump_end_position: whether to bump the end position by 1 or not (this is useful for interval trees and interval lists)
        :param chr_rep_as_int: whether to represent the chromosome as an int or not (this is useful for interval trees and interval lists)
        """
        # load from file
        if isinstance(maf_file, str):
            self.maf_file = maf_file
            self.col_positions = extract_maf_col_positions(maf_file)
            self.backed = backed
            self.mafs: List[SNP] = []

            # Open function depending on file type
            open_func = gzip.open if is_gzipped(maf_file) else open
            mode = "rt" if is_gzipped(maf_file) else "r"

            if backed:
                self.mafs = None
                # https://stackoverflow.com/a/32607817/13175187
                with open_func(self.maf_file, mode) as file:
                    self.length = (
                        sum(1 for line in file if line.strip()) - 1
                    )  # subtract 1 for header
            else:
                with open_func(maf_file, mode) as f:
                    # skip header
                    lines = f.readlines()[1:]
                    for line in lines:
                        # some bed files have more than 3 columns, so we just take the first 3
                        line = line.strip().split(MAF_FILE_DELIM)
                        self.mafs.append(
                            SNP(
                                hugo_symbol=self._extract_value_from_col(
                                    MAF_HUGO_SYMBOL_COL_NAME, line
                                ),
                                entrez_gene_id=self._extract_value_from_col(
                                    MAF_ENTREZ_GENE_ID_COL_NAME, line
                                ),
                                center=self._extract_value_from_col(MAF_CENTER_COL_NAME, line),
                                ncbi_build=self._extract_value_from_col(
                                    MAF_NCBI_BUILD_COL_NAME, line
                                ),
                                chromosome=self._extract_value_from_col(
                                    MAF_CHROMOSOME_COL_NAME, line
                                ),
                                start_position=int(
                                    self._extract_value_from_col(MAF_START_COL_NAME, line)
                                ),
                                end_position=int(
                                    self._extract_value_from_col(MAF_END_COL_NAME, line)
                                ),
                                strand=self._extract_value_from_col(MAF_STRAND_COL_NAME, line),
                            )
                        )
                    self.length = len(self.mafs)

                # post process according to flags
                for maf in self.mafs:
                    if bump_end_position:
                        maf.end_position += 1
                    if not chr_rep_as_int:
                        maf.chromosome = "chr" + str(maf.chromosome)
        else:
            raise ValueError("mafs must be a path to a maf file")

    def __len__(self):
        return self.length

    def __getitem__(self, key):
        if self.backed:
            raise NotImplementedError("Backed MAFs do not currently support indexing.")
        else:
            return self.mafs[key]

    def __iter__(self):
        if self.backed:
            # Open function depending on file type
            open_func = gzip.open if is_gzipped(self.maf_file) else open
            mode = "rt" if is_gzipped(self.maf_file) else "r"

            with open_func(self.maf_file, mode) as f:
                # skip header
                for i, line in enumerate(f):
                    if i == 0:
                        continue
                    line = line.strip().split(MAF_FILE_DELIM)
                    yield SNP(
                        hugo_symbol=self._extract_value_from_col(MAF_HUGO_SYMBOL_COL_NAME, line),
                        entrez_gene_id=self._extract_value_from_col(
                            MAF_ENTREZ_GENE_ID_COL_NAME, line
                        ),
                        center=self._extract_value_from_col(MAF_CENTER_COL_NAME, line),
                        ncbi_build=self._extract_value_from_col(MAF_NCBI_BUILD_COL_NAME, line),
                        chromosome=self._extract_value_from_col(MAF_CHROMOSOME_COL_NAME, line),
                        start_position=self._extract_value_from_col(MAF_START_COL_NAME, line),
                        end_position=self._extract_value_from_col(MAF_END_COL_NAME, line),
                        strand=self._extract_value_from_col(MAF_STRAND_COL_NAME, line),
                    )

        else:
            for maf in self.mafs:
                yield maf

    def __repr__(self):
        return f"MAF({self.maf_file})"


# TODO: This belongs somewhere else; does it even make sense?
class TokenizedRegionSet(object):
    """Represents a tokenized region set"""

    def __init__(self, tokens: np.ndarray, universe: RegionSet):
        self.tokens = tokens
        self.universe = universe


# Write a class representing a collection of RegionSets
# TODO: This shouldn't read in the actual files, it should just represent the files and use lazy loading
class RegionSetCollection(object):
    """Represents a collection of RegionSets"""

    def __init__(self, region_sets: List[RegionSet] = None, file_globs: List[str] = None):
        if region_sets:
            self.region_sets = region_sets
        elif file_globs:
            self.region_sets = []
            for glob in file_globs:
                self.region_sets.extend([RegionSet(path) for path in glob.glob(glob)])

    def __getitem__(self, key):
        return self.region_sets[key]

    def __len__(self):
        return len(self.region_sets)


# Do we need an EmbeddingSet class?
class EmbeddingSet(object):
    """Represents embeddings and labels"""

    embeddings: np.ndarray
    labels: list



from . import FileTokenizer


class BEDToolsTokenizer(FileTokenizer):
    """A tokenizer that uses bedtools to tokenize BED files"""

    def __init__(self, bedtools_path: str, universe_path: str = None):
        """Initialize a BEDToolsTokenizer

        Args:
            bedtools_path (str): Path to a bedtools binary.
            universe_path (str): Path to a universe BED file.
        """
        self.bedtools_path = bedtools_path
        self.universe_path = universe_path

    def tokenize(self, input_globs: list[str], universe_path: str = None) -> RegionSet:
        """Tokenize a RegionSet using bedtools"""

        universe_path = universe_path or self.universe_path

        # loop through globs and tokenize each file
        for glob in input_globs:
            for path in glob.glob(glob):
                _tokenize_one(path, universe_path)

    def _tokenize_one(self, input_path: str, universe_path: str):
        output_path = os.path.join(input_path, "tokenized.bed")
        bedtools_path = self.bedtools_path
        # bedtools can't actually read from stdin, so we have to use a temporary file...

        # sort_process = subprocess.Popen(shlex.split(f"sort -k1,1V -k2,2n {input_path}"), stdout=subprocess.PIPE)
        # bedtools_process = subprocess.Popen(
        #     shlex.split(f"{bedtools_path} intersect -a {universe} -b  -u -f {fraction}"),
        #     stdin = sort_process.stdout,
        #     stdout = output_file,
        # )
        # bedtools_process.communicate()

        # get a temporary file path using tempfile
        import templfile

        with tempfile.NamedTemporaryFile() as temp_path, open(output_path, "w") as output_file:
            # sort the input file
            sort_process = subprocess.Popen(
                shlex.split(f"sort -k1,1V -k2,2n {input_path}"), stdout=temp_path
            )
            sort_process.communicate()
            # tokenize the sorted file
            bedtools_process = subprocess.Popen(
                shlex.split(
                    f"{bedtools_path} intersect -a {universe} -b {temp_path} -u -f {fraction}"
                ),
                stdout=output_file,
            )
            bedtools_process.communicate()



import logging
from typing import List, Tuple

try:
    import torch
    import torch.nn as nn
    from rich.progress import track
    from torch.utils.data import Dataset
except ImportError:
    raise ImportError(
        "Please install Machine Learning dependencies by running 'pip install geniml[ml]'"
    )

from .const import DEFAULT_N_SHUFFLES, DEFAULT_NS_POWER, DEFAULT_WINDOW_SIZE, MODULE_NAME
from .utils import shuffle_documents

_LOGGER = logging.getLogger(MODULE_NAME)


class Region2VecDataset(Dataset):
    def __init__(self, samples: List[Tuple[List[any], any]]):
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx) -> Tuple[List[any], any]:
        # we need to return things as a tensor for proper batching
        return self.samples[idx]


def generate_window_training_data(
    data: List[List[any]],
    window_size: int = DEFAULT_WINDOW_SIZE,
    n_shuffles: int = DEFAULT_N_SHUFFLES,
    threads: int = None,
    padding_value: any = 0,
    return_tensor: bool = True,
) -> List[Tuple[List[any], any]]:
    """
    Generates the windowed training data by sliding across the region sets. This is for the CBOW model.

    :param List[any] data: The data to generate the training data from.
    :param int window_size: The window size to use.
    :param int n_shuffles: The number of shuffles to perform.
    :param int threads: The number of threads to use.
    :param any padding_value: The padding value to use.
    :param bool return_tensor: Whether or not to return the data as a tensor.

    :return Tuple[List[List[any]], List[any]]: The contexts and targets.
    """
    _LOGGER.info("Generating windowed training data.")

    # shuffle the documents
    documents = shuffle_documents(
        [[t for t in tokens] for tokens in data], n_shuffles=n_shuffles, threads=threads
    )

    # compute the context length (inputs)
    context_len_req = 2 * window_size
    # contexts = []
    # targets = []
    samples = []
    for document in track(documents, total=len(documents), description="Generating training data"):
        for i, target in enumerate(document):
            context = document[max(0, i - window_size) : i] + document[i + 1 : i + window_size + 1]

            # pad the context if necessary
            if len(context) < context_len_req:
                context = context + [padding_value] * (context_len_req - len(context))

            # contexts.append(context)
            # targets.append(target)
            if return_tensor:
                samples.append(
                    (
                        torch.tensor(context, dtype=torch.long),
                        torch.tensor(target, dtype=torch.long),
                    )
                )
            else:
                samples.append((context, target))

    # return contexts, targets
    return samples


def generate_window_training_data_wrap(
    data: List[List[any]],
    window_size: int = DEFAULT_WINDOW_SIZE,
    n_shuffles: int = DEFAULT_N_SHUFFLES,
    threads: int = None,
    padding_value: any = 0,
    return_tensor: bool = True,
) -> List[Tuple[List[any], any]]:
    """
    Generates the windowed training data by sliding across the region sets. When the sliding window runs into the bounds of the list, it wraps around to the start or end of the array.

    :param List[any] data: The data to generate the training data from.
    :param int window_size: The window size to use.
    :param int n_shuffles: The number of shuffles to perform.
    :param int threads: The number of threads to use.
    :param any padding_value: The padding value to use.
    :param bool return_tensor: Whether or not to return the data as a tensor.

    :return Tuple[List[List[any]], List[any]]: The contexts and targets.
    """
    _LOGGER.info("Generating windowed training data.")

    # shuffle the documents
    documents = shuffle_documents(
        [[t for t in tokens] for tokens in data], n_shuffles=n_shuffles, threads=threads
    )

    samples = []

    for document in track(documents, total=len(documents), description="Generating training data"):
        for i in range(0, window_size):
            target = document[i]
            context = (
                document[i - window_size :] + document[0:i] + document[i + 1 : i + 1 + window_size]
            )
            if return_tensor:
                samples.append(
                    (
                        torch.tensor(context, dtype=torch.long),
                        torch.tensor(target, dtype=torch.long),
                    )
                )
            else:
                samples.append((context, target))
        for i in range(window_size, len(document) - window_size):
            target = document[i]
            context = document[i - window_size : i] + document[i + 1 : i + 1 + window_size]
            if return_tensor:
                samples.append(
                    (
                        torch.tensor(context, dtype=torch.long),
                        torch.tensor(target, dtype=torch.long),
                    )
                )
            else:
                samples.append((context, target))
        for i in range(len(document) - window_size, len(document)):
            target = document[i]
            context = (
                document[i - window_size : i]
                + document[i + 1 :]
                + document[0 : i - len(document) + window_size + 1]
            )
            if return_tensor:
                samples.append(
                    (
                        torch.tensor(context, dtype=torch.long),
                        torch.tensor(target, dtype=torch.long),
                    )
                )
            else:
                samples.append((context, target))

    return samples


def generate_frequency_distribution(tokens: List[List[int]], vocab_length: int) -> torch.Tensor:
    """
    Generate the frequency distribution of the tokens.

    :param List[List[int]] tokens: The tokens to generate the frequency distribution from.
    """
    tokens_flat = [t for tokens in tokens for t in tokens]

    # create a tensor of all zeros with the length of the vocabulary
    freq_dist = torch.zeros(vocab_length, dtype=torch.float)

    # count the number of times each token appears
    for token in track(
        tokens_flat,
        total=len(tokens_flat),
        description="Generating frequency distribution",
    ):
        freq_dist[token] += 1

    # normalize the frequency distribution
    freq_dist /= freq_dist.sum()

    return freq_dist


class NegativeSampler:
    def __init__(
        self,
        freq_dist: torch.Tensor,
        power: float = DEFAULT_NS_POWER,
        batch_size: int = None,
    ):
        """
        Initialize the negative sampler.

        :param torch.Tensor freq_dist: List of frequencies for each token. Must be normalized.
        :param float power: The power to use for the negative sampling. It is not recommended to change this.
        """
        self.dist = freq_dist**power
        self.dist /= self.dist.sum()
        self.power = power
        self.batch_size = batch_size

    def sample(self, k: int = 5, batch_size: int = None) -> torch.Tensor:
        """
        Sample from the negative sampler.

        :param int k: The number of samples to draw.
        """
        batch_size = batch_size or self.batch_size
        if batch_size is None:
            raise ValueError(
                "Must provide batch_size to sample from negative sampler. This can be set in the constructor or in the sample method."
            )
        negative_samples = torch.multinomial(self.dist, batch_size * k, replacement=True)
        return negative_samples.view(batch_size, k)


class NegativeSampleDataset(Dataset):
    def __init__(self, samples: torch.Tensor):
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx) -> torch.Tensor:
        return self.samples[idx]


# negative sampling loss
class NSLoss(nn.Module):
    def __init__(self):
        super(NSLoss, self).__init__()

    def forward(
        self,
        context: torch.Tensor,
        negative_samples: torch.Tensor,
        target: torch.Tensor,
    ):
        """
        :param torch.Tensor context: The context vectors.
        :param torch.Tensor negative_samples: The negative sample vectors.
        :param torch.Tensor target: The target vectors.
        """
        # there is one target that gets mapped to each context
        target_v_context = target.unsqueeze(1).expand(
            context.shape[0], context.shape[1], context.shape[2]
        )
        target_v_neg = target.unsqueeze(1).expand(
            negative_samples.shape[0],
            negative_samples.shape[1],
            negative_samples.shape[2],
        )

        # target is now of shape (batch_size, num_context_vectors, embedding_size)
        # negative_samples is of shape (batch_size, num_negative_samples, embedding_size)
        # context is of shape (batch_size, num_context_vectors, embedding_size)

        # compute the dot product between the context and target
        pos_loss = torch.sum(
            torch.nn.functional.logsigmoid(torch.bmm(context, target_v_context.transpose(1, 2)))
        )
        neg_loss = torch.sum(
            torch.nn.functional.logsigmoid(
                torch.bmm(-negative_samples, target_v_neg.transpose(1, 2))
            )
        )

        return -(pos_loss + neg_loss)



from typing import Literal

MAF_HUGO_SYMBOL_COL_NAME = "Hugo_Symbol"
MAF_ENTREZ_GENE_ID_COL_NAME = "Entrez_Gene_Id"
MAF_CENTER_COL_NAME = "Center"
MAF_NCBI_BUILD_COL_NAME = "NCBI_Build"
MAF_CHROMOSOME_COL_NAME = "Chromosome"
MAF_START_COL_NAME = "Start_position"
MAF_END_COL_NAME = "End_position"
MAF_STRAND_COL_NAME = "Strand"

MAF_COLUMN = Literal[
    "Hugo_Symbol",
    "Entrez_Gene_Id",
    "Center",
    "NCBI_Build",
    "Chromosome",
    "Start_Position",
    "End_Position",
    "Strand",
]

MAF_FILE_DELIM = "\t"



import os
import sys

import logmuse
from gensim.models import Word2Vec

from ._version import __version__
from .argparser import build_argparser
from .const import *
from .main import convert_anndata_to_documents, load_scanpy_data, shuffle_documents, train


def main():
    """MAIN"""
    parser = logmuse.add_logging_options(build_argparser())
    args, _ = parser.parse_known_args()

    if args.input is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    global _LOGGER
    _LOGGER = logmuse.logger_via_cli(args, make_root=True)
    _LOGGER.debug(f"versions: scembed {__version__}")
    _LOGGER.debug(f"Args: {args}")

    # TODO: store parameters in a file, not in the filenames?
    # TODO: load a config file for all the parameters instead of specifying

    try:
        os.makedirs(args.output, exist_ok=True)
    except OSError as error:
        _LOGGER.error(error)
        sys.exit(1)

    try:
        os.makedirs(os.path.join(args.output, "figs"), exist_ok=True)
    except OSError as error:
        _LOGGER.error(error)
        sys.exit(1)

    _LOGGER.info("Loading data")

    # load AnnData object
    sc_data = load_scanpy_data(args.input)

    # re-create documents dictionary from AnnData Object
    documents = convert_anndata_to_documents(sc_data)

    if args.model:
        model = Word2Vec.load(args.model)
        _LOGGER.info(f"Loaded {args.model}")
    else:
        model_name = (
            f"_nocells{args.nocells}_noreads{args.noreads}_"
            f"dim{args.dimension}_win{args.window_size}_"
            f"mincount{args.min_count}_"
            f"shuffle{args.shuffle_repeat}.model"
        )
        model_filename = os.path.join(args.output, args.title + model_name)
        _LOGGER.info(f"Shuffling documents")
        shuffeled_documents = shuffle_documents(documents, int(args.shuffle_repeat))
        _LOGGER.info(f"Constructing model")
        model = train(
            shuffeled_documents,
            window_size=int(args.window_size),
            dim=int(args.dimension),
            min_count=int(args.min_count),
            nothreads=int(args.nothreads),
        )
        model.save(model_filename)
        _LOGGER.info(f"Model saved as: {model_filename}")

    _LOGGER.info(f"Number of words in w2v model: {len(model.wv.vocab)}")
    _LOGGER.info(f"Pipeline Complete!")



import os
import warnings
from logging import getLogger

import numpy as np
import pandas as pd

from ..const import PKG_NAME
from .distance import run_distance
from .intersection import run_intersection
from .likelihood import hard_universe_likelihood, likelihood_flexible_universe
from .utils import check_if_uni_flexible

_LOGGER = getLogger(PKG_NAME)


def run_all_assessment_methods(
    raw_data_folder,
    file_list,
    universe,
    no_workers,
    folder_out,
    pref,
    save_each,
    overlap=False,
    distance_f_t_u=False,
    distance_f_t_u_flex=False,
    distance_u_t_f=False,
    distance_u_t_f_flex=False,
):
    """
    Assess universe fit to collection using overlap and distance metrics
    :param str raw_data_folder: path to raw files from the collection
    :param str file_list: path to file with list of files in the collection
    :param str universe: path to universe that is being assessed
    :param int no_workers: number of workers for multiprocessing
    :param str folder_out: output folder
    :param str pref: prefixed used for creating output files
    :param bool save_each: if save output of distance metrics for each region
    :param bool overlap: if calculate overlap metrics
    :param bool distance_f_t_u: if calculate distance from file to universe metrics
    :param bool distance_f_t_u_flex: if calculate flexible distance from file to universe metrics
    :param bool distance_u_t_f: if calculate distance from universes to file metrics
    :param bool distance_u_t_f_flex: if calculate flexible distance from universes to file metrics
    """
    if not any(
        [
            overlap,
            distance_f_t_u,
            distance_f_t_u_flex,
            distance_u_t_f,
            distance_u_t_f_flex,
        ]
    ):
        raise AttributeError("Choose at least one assessment method")
    if not any(
        [
            distance_f_t_u,
            distance_f_t_u_flex,
            distance_u_t_f,
            distance_u_t_f_flex,
        ]
    ):
        warnings.warn("Unused argument: save_each")
    asses_results = []
    if overlap:
        r_overlap = run_intersection(
            raw_data_folder,
            file_list,
            universe,
            no_workers,
        )
        r_overlap.columns = [
            "file",
            "univers/file",
            "file/universe",
            "universe&file",
        ]
        asses_results.append(r_overlap)
        _LOGGER.info("DONE: Overlap")
    if distance_f_t_u:
        r_distance = run_distance(
            raw_data_folder,
            file_list,
            universe,
            no_workers,
            False,
            folder_out,
            pref + "_dist_file_to_universe",
            save_each,
            False,
        )
        r_distance.columns = ["file", "median_dist_file_to_universe"]
        asses_results.append(r_distance)
        _LOGGER.info("DONE: Distance file to universe")
    if distance_f_t_u_flex:
        r_distance_flex = run_distance(
            raw_data_folder,
            file_list,
            universe,
            no_workers,
            True,
            folder_out,
            pref + "_dist_file_to_universe_flex",
            save_each,
            False,
        )
        r_distance_flex.columns = ["file", "median_dist_file_to_universe_flex"]
        asses_results.append(r_distance_flex)
        _LOGGER.info("DONE: Flexible distance file to universe")

    if distance_u_t_f:
        r_distance_utf = run_distance(
            raw_data_folder,
            file_list,
            universe,
            no_workers,
            False,
            folder_out,
            pref + "_dist_universe_to_file",
            save_each,
            True,
        )
        r_distance_utf.columns = ["file", "median_dist_universe_to_file"]
        asses_results.append(r_distance_utf)
        _LOGGER.info("DONE: Distance universe to file")
    if distance_u_t_f_flex:
        r_distance_utf_flex = run_distance(
            raw_data_folder,
            file_list,
            universe,
            no_workers,
            True,
            folder_out,
            pref + "median_dist_universe_to_file_flex",
            save_each,
            True,
        )
        r_distance_utf_flex.columns = [
            "file",
            "median_dist_universe_to_file_flex",
        ]
        asses_results.append(r_distance_utf_flex)
        _LOGGER.info("DONE: Flexible distance universe to file")

    df = asses_results[0]
    for i in asses_results[1:]:
        df = pd.merge(df, i, on="file")
    df.to_csv(os.path.join(folder_out, pref + "_data.csv"), index=False)


def get_rbs(f_t_u, u_t_f):
    """
    Calculate RBS
    """
    a = 101 / (f_t_u + 100)
    b = 101 / (u_t_f + 100)
    rbs = (10 * a + b) / 11
    return rbs


def get_mean_rbs(folder, file_list, universe, no_workers, flexible=False):
    """
    Calculate average RBS of the collection
    :param str folder: path to folder with the collection
    :param str file_list: path to file with list of files in the collection
    :param str universe: path to the universe
    :param int no_workers: number of workers for multiprocessing
    :param bool flexible: if to calculate flexible version of the metric
    :return int: average RBS
    """
    file_to_uni = run_distance(
        folder,
        file_list,
        universe,
        no_workers,
        flexible=flexible,
        uni_to_file=False,
    )

    uni_to_file = run_distance(
        folder,
        file_list,
        universe,
        no_workers,
        flexible=flexible,
        uni_to_file=True,
    )
    rbs = get_rbs(file_to_uni[1], uni_to_file[1])
    return np.mean(rbs)


def get_rbs_from_assessment_file(file, cs_each_file=False, flexible=False):
    """
    Calculate RBS form file with results of metrics per file
    :param str file: path to file with assessment results
    :param bool cs_each_file: if report RBS for each file, not average for the collection
    :param bool flexible: if use flexible version of the metric
    """
    df = pd.read_csv(file, index_col=(0))
    if flexible:
        df["f_t_u"] = df["median_dist_file_to_universe_flex"]
        df["u_t_f"] = df["median_dist_universe_to_file_flex"]
    else:
        df["f_t_u"] = df["median_dist_file_to_universe"]
        df["u_t_f"] = df["median_dist_universe_to_file"]
    df["RBS"] = get_rbs(df["f_t_u"], df["u_t_f"])
    if cs_each_file:
        return df
    else:
        return df["RBS"].mean()


def get_f_10_score(
    folder,
    file_list,
    universe,
    no_workers,
):
    """
    Get F10 score for a universes and collection of files
    :param str folder: path to folder with the collection
    :param str file_list: path to file with list of files in the collection
    :param str universe: path to the universe
    :param int no_workers: number of workers for multiprocessing
    :return int: average F10 score
    """
    res = run_intersection(
        folder,
        file_list,
        universe,
        no_workers,
    )
    res = np.array(res)
    res = res[:, 1:]
    res = res.astype("float")
    recall = res[:, 2] / (res[:, 2] + res[:, 1])
    precision = res[:, 2] / (res[:, 2] + res[:, 0])
    f_10 = (1 + 10**2) * (precision * recall) / ((10**2 * precision) + recall)
    return np.mean(f_10)


def get_f_10_score_from_assessment_file(file, f10_each_file=False):
    """
    Get F10 score from assessment output file
    :param str file: path to file with assessment results
    :param bool f10_each_file: if report F10 for each file, not average for the collection
    """
    df = pd.read_csv(file, index_col=(0))
    r = df["universe&file"] / (df["universe&file"] + df["file/universe"])
    p = df["universe&file"] / (df["universe&file"] + df["univers/file"])
    df["F_10"] = (1 + 10**2) * (p * r) / ((10**2 * p) + r)
    if f10_each_file:
        return df["F_10"]
    else:
        return df["F_10"].mean()


def get_likelihood(
    model_file,
    universe,
    cove_folder,
    cove_prefix="all",
    flexible=False,
    save_peak_input=False,
):
    """
    Calculate universe likelihood given collection
    :param str model_file: path to file with likelihood model
    :param str universe: path to the universe
    :param str cove_folder: path to the coverage folder
    :param str cove_prefix: prefixed used for generating coverage
    :param bool flexible: if to calculate flexible likelihood
    :param bool save_peak_input:  if to save likelihood input of each region
    :return:
    """
    if flexible:
        lh = likelihood_flexible_universe(
            model_file, universe, cove_folder, cove_prefix, save_peak_input
        )
    else:
        if save_peak_input:
            warnings.warn("Unused argument: save_peak_input")
        lh = hard_universe_likelihood(model_file, universe, cove_folder, cove_prefix)

    return lh


def filter_universe(
    universe,
    universe_filtered,
    min_size=0,
    min_coverage=0,
    filter_lh=False,
    model_file=None,
    cove_folder=None,
    cove_prefix=None,
    lh_cutoff=0,
):
    """
    Filter universe by region size, coverage by collection, likelihood
    :param str universe: path to input universe
    :param str universe_filtered: path to output filtered universe
    :param int min_size: minimum size of the region in the output universe
    :param int min_coverage: minimum number coverage of universe region by collection
    :param bool filter_lh: if use likelihood to filter universe
    :param str model_file: path to collection likelihood model
    :param str cove_folder: path to folder with coverage tracks
    :param str cove_prefix: prefixed used for creating tracks
    :param int lh_cutoff: minimum likelihood input
    """
    if filter_lh:
        check_if_uni_flexible(universe)
        if not all([model_file, cove_folder, cove_prefix]):
            miss_args = []
            if not model_file:
                miss_args.append("model_file")
            if not cove_folder:
                miss_args.append("cove_folder")
            if not cove_prefix:
                miss_args.append("cove_prefix")
            raise ValueError(
                "Missing {} for peak likelihood calculations.".format(",".join(miss_args))
            )
        likelihood_flexible_universe(model_file, universe, cove_folder, cove_prefix, True)
        universe = universe + "_peak_likelihood"
    with open(universe) as uni:
        with open(universe_filtered, "w+") as uni_flt:
            for i in uni:
                j = i.split("\t")
                j[1], j[2], j[4] = int(j[1]), int(j[2]), int(j[4])
                if j[2] - j[1] > min_size:
                    if j[4] > min_coverage:
                        if filter_lh:
                            if float(j[9].strip("\n")) > lh_cutoff:
                                uni_flt.write(i)
                        else:
                            uni_flt.write(i)



def build_subparser(parser):
    """
    Builds argument parser.
    """

    parser.add_argument(
        "--model-file",
        help="path to file with lh model",
        required=True,
        type=str,
    )
    parser.add_argument("--file-no", help="number of files used to make the model", type=int)
    parser.add_argument(
        "--coverage-folder",
        help="path to coverage folder",
        required=True,
        type=str,
    )
    parser.add_argument(
        "--coverage-prefix",
        help="prefix used when making coverage files",
        default="all",
        type=str,
    )
    parser.add_argument(
        "--force",
        help="if overwrite existing model",
        action="store_true",
    )
    return parser
